import os
import traceback
from typing import List, Tuple

import PIL
from PIL import Image

from pho_mos.mosaic import Rectangle, Point
from pho_mos.mosaic import get_average_color, get_next_near_image, add_colour_filter, \
    calc_gradient_field
from .preprocessing import resize_img, open_as_pil_image, square_img

from pho_mos.util import get_logger

Image.MAX_IMAGE_PIXELS = None
LOGGER = get_logger("MOSAIC")


class MosiaicSourceImage:

    def __init__(self, path, cell_size: int, priority: int = 1):
        # перегоняем в PIL, заготавливаем маленький квадрат, считаем средний цвет
        self.origin_img: PIL.Image = open_as_pil_image(path)
        self.squared_image: PIL.Image = square_img(self.origin_img, cell_size)
        self.image_average_color = get_average_color(self.squared_image)

        self.priority = priority
        self.is_used = False

# для сортировки в списке
def get_mosaic_image_priority(image: MosiaicSourceImage):
    return image.priority

'''
рисует мозайку попиксельно заменяя на картинку
'''
def build_mosaic(target_image: PIL.Image.Image,
                 source_images: List[MosiaicSourceImage],
                 ukrup_field_max_percent: float = 0.1,
                 cell_size: int = 64,
                 not_ukrup_fields: List[Tuple] = []
                 ) -> PIL.Image.Image:
    """
    Build mosaic from target_image and list of source_images.

    NOTE:
        build картинки дб приведены к квадрату ребром img_size;

    :param not_ukrup_fields:
    :param ukrup_field_max_percent:
    :param target_image: Image
    :param source_images: List[str] - список к путями к изображениям - кирпичам
    :param random_param: randomization parameter to choose best image for pixel;
    :param cell_size: размер сетки - минимальный размер картинки
    :return:
    """
    try:
        LOGGER.debug("BUILD START. target_image: " + str(target_image.width)+"x"+str(target_image.height))
        LOGGER.debug("BUILD START. building_images: " + str(source_images))
        target_image_pixels = target_image.load()

        # сортируем картинки по приоритету
        source_images.sort(key=get_mosaic_image_priority)
        # TODO избавиться от этого:
        source_image_average_colors = {}
        for index, img in enumerate(source_images):
            source_image_average_colors[index] = img.image_average_color

        output_dims = (
            target_image.width * cell_size,
            target_image.height * cell_size
        )

        output_image = Image.new("RGB", output_dims)
        # область, хранящая прямоугольники укрупнения
        output_rectangle: Rectangle = Rectangle(target_image.width, target_image.height)
        LOGGER.debug("Created new output image " + str(output_image.width)+"x"+str(output_image.height))

        # словарь с полем градиентов в точках
        target_image_gradient_field_dict = calc_gradient_field(target_image)

        # прямоугольник, содержащий прямоугольники, соответствующие зонам, которые не нужно укрупнять
        not_ukr_fields_rectangle: Rectangle = Rectangle(target_image.width, target_image.height)
        for field in not_ukrup_fields:
            start_point: Point = Point(field[0]*target_image.width, field[1]*target_image.height)
            rectangle: Rectangle = Rectangle((field[2]-field[0])*target_image.width,
                                             (field[3]-field[1])*target_image.height)
            not_ukr_fields_rectangle.add_inner_rectangle(start_point, rectangle)

        # минимальный и максимальный размер области укрупнения
        ukrup_field_max_size: int = int(min(target_image.width, target_image.height) * ukrup_field_max_percent)
        LOGGER.debug("Max ukrup field size "+str(ukrup_field_max_size))

        max_gradient_porog = 120

        LOGGER.debug("RENDER START")

        # заполняем output_rectangle
        # сквозной проход квадратами от размеров ukrup_field_max_size до >3 (выведено эмпирически)
        step = ukrup_field_max_size
        shape_sizes = []

        while step > 3:
            shape_sizes.append(step)
            step = int(step*0.8)

        for size in shape_sizes:

            LOGGER.debug("*** Check size "+str(size))
            x_start_shape = 0

            # цикл по всем возможным точкам начала фигуры
            while x_start_shape <= target_image.width-size:
                y_start_shape = 0
                while y_start_shape <= target_image.height-size:

                    # собираем список градиентов текущего квадрата
                    is_empty_rect_field = True
                    is_consist_greater_porog_gradient_point = False

                    for i in range(0, size-1):
                        # если внутри квадрата уже ловили занятую точку, заканчиваес цикл в холостую
                        if not is_empty_rect_field:
                            break

                        for j in range(0, size-1):
                            # если внутри квадрата уже ловили занятую точку, заканчиваес цикл в холостую
                            if not is_empty_rect_field:
                                break

                            # если точка внутри квадрата занята, пропускаем квадрат
                            if (output_rectangle.is_occupied_point(point=Point(x_start_shape+i, y_start_shape+j)) or
                                not_ukr_fields_rectangle.is_occupied_point(point=Point(x_start_shape+i, y_start_shape+j))):
                                is_empty_rect_field = False
                                break

                            # проверяем, содержит ли область точку с большим градиентом
                            if target_image_gradient_field_dict[(x_start_shape+i, y_start_shape+j)]>max_gradient_porog:
                                is_consist_greater_porog_gradient_point = True


                    # если внутри квадрата уже ловили занятую точку, пропускаем ход и идем к след.квадрату
                    if not is_empty_rect_field:
                        # LOGGER.debug("Field not empty")
                        y_start_shape = y_start_shape + int(size/4)
                        continue

                    # если в области нет точки с высоким градиентом, заполняем изображением size
                    if not is_consist_greater_porog_gradient_point:

                        # заполнение укрупненной области фигурой размера size
                        new_rectangle: Rectangle = Rectangle(size, size)
                        # достаем наиболее подходящую картинку (пока по центральному пикселю)
                        best_img: MosiaicSourceImage  = get_next_near_image(
                            target_image_pixels[int(x_start_shape+size/2), int(y_start_shape+size/2)],
                            source_images,
                            with_used=False,
                            last_use=True,
                            use_priority=True
                        )
                        if best_img is None:
                            raise Exception("Not found best image!")

                        new_rectangle.image = add_colour_filter(target_image_pixels[x_start_shape, y_start_shape],
                                                                best_img.origin_img)
                        new_rectangle.image = square_img(new_rectangle.image, size * cell_size)

                        output_rectangle.add_inner_rectangle(Point(x_start_shape, y_start_shape), new_rectangle)

                        LOGGER.debug("Draw rectangle " + str(x_start_shape) + ":" + str(y_start_shape) + " " + str(new_rectangle.image))
                        output_image.paste(new_rectangle.image, (x_start_shape * cell_size, y_start_shape * cell_size))

                        # если квадрат добавлен, двигаемся сразу к след.точке
                        y_start_shape = y_start_shape + int(size)
                    else:
                        # LOGGER.debug("More than porog")
                        y_start_shape = y_start_shape + int(size/4)

                # по оси абцисс двигаемся скольжением по size / 2
                x_start_shape = x_start_shape + int(size/4)


        #  заполняем все незаполненные поля фигурами размера size
        for x in range(0, target_image.width):
            for y in range(0, target_image.height):
                # если точка занята, не трогаем
                if output_rectangle.is_occupied_point(point=Point(x, y)):
                    continue

                best_img: MosiaicSourceImage = get_next_near_image(
                    target_image_pixels[x, y],
                    source_images,
                    with_used=True,
                    use_priority=False
                )
                img = add_colour_filter(target_image_pixels[x, y],
                                        best_img.squared_image)
                output_image.paste(
                    img, (x * cell_size, y * cell_size)
                )


        LOGGER.debug("BUILD END")

        return output_image

    except Exception as exception:
        LOGGER.debug("Exception in build_mosaic: {0}".format(str(traceback.format_exc())))



'''
создаем мозайку с укрупнением
'''
def create_mosaic(img_path: str,
                  source_dirs: List[str],
                  target_path: str,
                  cell_size: int = 64,
                  resize_factor: float = 0.1,
                  ukrup_field_max_percent: float = 0.1,
                  not_ukrup_fields: List[Tuple] = []
                  ) -> None:
    """
    создать укрупненную мозайку

    :param img_path:
    :param source_dirs: список директорий, из которых брать элементы, в порядке приоритета!
    :param target_path:
    :param cell_size: - размер ячейки, на которые производится разбиение
    :param resize_factor: - во сколько раз увеличится изображение (для больших фото уменьшаем)
    :param ukrup_field_min_percent - минимальный размер области которую можно укрупнить
    :param not_ukrup_fields -  список областей, в которых не требуется укрупнение. область в формате процентов от x,y..
    :return:
    """
    try:

        LOGGER.debug("IN create_mosaic: INSIDE")

        # сделали resize
        target_image: PIL.Image = resize_img(
            img_path,
            resize_factor=resize_factor,
            is_save=False
        )

        if target_image is None:
            LOGGER.error("Cannot open target image! Check path!!!")
            raise Exception("No image under path {0} ".format(img_path))

        LOGGER.debug("IN create_mosaic: PROCESS TARGET")

        # достаем картинки, из которых строится изображение и режем до квадратных
        source_images: List[MosiaicSourceImage] = []
        for index, source_dir in enumerate(source_dirs):
            for name in os.listdir(source_dir):
                path = os.path.join(source_dir, name)
                LOGGER.debug("Append source image: "+path+"; priority "+str(index))
                img: MosiaicSourceImage = MosiaicSourceImage(path, cell_size, priority=index)
                source_images.append(img)

        LOGGER.debug("IN create_mosaic: PROCESS build_batch_images {0}".format(source_images))
        LOGGER.debug("IN create_mosaic: START BUILD")

        image = build_mosaic(target_image, source_images,
                             ukrup_field_max_percent = ukrup_field_max_percent,
                             cell_size= cell_size, not_ukrup_fields = not_ukrup_fields)
        if image is not None:
            image.save(target_path)

    except Exception as e:
        LOGGER.debug("Exception in create_mosaic: {0}".format(traceback.format_exc()))



import hashlib
import os.path

BLOCK_SIZE = 512

# Мб пригодится в будущем хешерование больших файлов из ОС:
def from_file_sha(file_path, blocksize=BLOCK_SIZE) -> str:
    sha1 = hashlib.sha1()
    with open(file_path, "rb") as f:
        buf = f.read(blocksize)
        while buf:
            buf = f.read(blocksize)
            if not buf:
                break
            sha1.update(buf)
    return sha1.hexdigest()



import concurrent.futures
import urllib.request as url_req
import os

from pho_mos.util import get_logger
LOGGER = get_logger("PARALLEL_UTIL")

def parallel_download(URLS: list,
                      result_path: str,
                      max_count: int = 1000,
                      image_prefix:str='img') -> bool:

    counter = 0
    # Retrieve a single page and report the URL and contents
    def load_url(img_link, timeout) -> bool:

        try:

            LOGGER.debug("Try download " + img_link)
            request_result = url_req.urlopen(img_link, timeout=timeout)
            raw_img = request_result.read()

            count_img_in_dir = len([i for i in os.listdir(result_path) if image_prefix in i]) + 1
            f = open(os.path.join(result_path, image_prefix + "_" + str(counter) + "." + 'jpeg'), 'wb')
            f.write(raw_img)
            f.close()
            LOGGER.debug("Success download")
            return True

        except Exception as e:
            LOGGER.debug("Could not load : " + str(img_link))
            LOGGER.debug(str(e))
            return False

    # We can use a with statement to ensure threads are cleaned up promptly
    with concurrent.futures.ThreadPoolExecutor(max_workers=len(URLS)) as executor:

        # Start the load operations and mark each future with its URL
        tasks = {executor.submit(load_url, url, 2): url for url in URLS}

        for future in concurrent.futures.as_completed(tasks):
            url = tasks[future]
            try:

                data = future.result()
                if data == True:
                    counter = counter + 1
                    if counter == max_count:
                        LOGGER.debug("Counter " + str(max_count) + " success download image")
                        executor.shutdown()
                        return True
            except Exception as exc:
                LOGGER.error('%r generated an exception: %s' % (url, exc))

        return True


import random

import PIL
from PIL import Image
from typing import List, Tuple
from collections import deque

from pho_mos.mosaic.transforms import RGBTransform
from pho_mos.util import get_logger

JPG_IMG_LEN = 3
PNG_IMG_LEN = 4
rgb_transform = RGBTransform()

Image.MAX_IMAGE_PIXELS = None
LOGGER = get_logger("MOSAIC_UTIL")


def add_colour_filter(_filter, image):
    return rgb_transform.mix_with(_filter, factor=.30).applied_to(image)


def get_colour_distance(target_colour, img_colour) -> int:
    """
    Count colour distance between two pixels.
    Манхетонское расстояние между двумя пикселями

    :param target_colour:
    :param img_colour:
    :return:
    """

    def _preprocess_img_len(img):
        if len(img) == JPG_IMG_LEN:
            return [col for col in img]  # + [255]  # jpg has no capacity
        elif len(img) == PNG_IMG_LEN:
            return img
        else:
            raise Exception(
                "Wrong img colour length size: {0}".format(len(img))
            )

    try:
        target_colour = _preprocess_img_len(target_colour)
        img_colour = _preprocess_img_len(img_colour)

        distance = 0
        for ind, colour in enumerate(target_colour):
            distance += abs(colour - img_colour[ind])
        return distance
    except Exception as exception:
        LOGGER.debug("Exception in get_average_colour: {0}".format(exception))


'''
считает средний свет картинки
'''


def get_average_color(image: PIL.Image) -> List[int]:
    """
    Counts average colour of the image.

    :param image: Image
    :return: List[int] size of image channels amount - список из трех цветов по (R,G,B,capacity).
    """
    try:
        pixels = image.load()
        channels_amount = len(pixels[0, 0])
        colours = [0] * channels_amount
        for x in range(image.width):
            for y in range(image.height):
                for channel in range(channels_amount):
                    colours[channel] += pixels[x, y][channel]

        pix_amount = image.width * image.height
        return list(
            map(lambda chanel: int(chanel / pix_amount), colours)
        )
    except Exception as exception:
        LOGGER.debug("Exception in get_average_colour: {0}".format(exception))


'''
возващает наиболее подходящую картинку
'''


def get_next_near_image(target_pixel: Tuple,
                        sorted_source_images: List[any],
                        with_used=True,
                        last_use=False,
                        use_priority=True) -> any:
    """
    Find nearest image from images_pixel with fixed accuracy.
    returns best MosiaicSourceImages

    :param with_used: использовать ли уже отданные картинки (с is_used = True)
    :param last_use: если True, выданная картинка больше не будет ранжироваться в следующих обращениях к get_best_image
    :param target_pixel: Tuple (R,G,B,capacity)
    :param accuracy: int
    :return:
    """
    try:
        min_distance = 1000
        result = sorted_source_images[-1]

        good_results_dequeue: deque = deque(maxlen=3)

        # в этой ссылке лучшая картинка, которая уже использовалась
        # она возвращается, если если все уже использованы
        # best_used_result = sorted_source_images[-1]

        # на сколько может быть больше расстояние между текущим результатом и новым, чтобы приоритет картинки позволил смениьть результат
        # параметр позволяет использовать более приоритетную картику, если она чуть дальше по расстоянию
        replace_distance_delta = 100

        for image in reversed(sorted_source_images):

            if not with_used and image.is_used:
                # пропускаем используемую картинку
                continue

            distance = get_colour_distance(target_pixel, image.image_average_color)
            if use_priority:
                # меняем если больше приоритет или меньше дистанция до точки
                if ((distance <= min_distance and image.priority <= result.priority) or
                        (distance <= min_distance + replace_distance_delta and image.priority < result.priority)):
                    result = image
                    min_distance = distance
                    good_results_dequeue.append(image)
            else:
                if distance <= min_distance:
                    result = image
                    min_distance = distance
                    good_results_dequeue.append(image)

        # если все использованы, а хотели новую, возвращаем использованную
        if len(good_results_dequeue) == 0 and not with_used:
            return get_next_near_image(target_pixel, sorted_source_images, with_used=True, use_priority=use_priority)

        for i in range(0, random.randrange(good_results_dequeue.__len__())):
            result = good_results_dequeue.popleft()

        if last_use:
            # проставляем, что уже использовалась
            result.is_used = True

        return result

    except Exception as exception:
        LOGGER.debug("Exception in get_best_image: {0}".format(exception))


'''
считает максимальную дельту между пикселями из массива и говорит, превышает ли она порог
'''


def check_max_pixel_delta_not_greater_porog(pixel_list: List[Tuple], porog: int) -> bool:
    """
    :param pixel_list: список элементов (R,G,B,capacity)
    :return: максимальное расстояние
    """
    max_distance = 0
    for i1, val1 in enumerate(pixel_list):
        for i2, val2 in enumerate(pixel_list):
            if i1 != i2:
                # LOGGER.debug("Calc colour distance between "+str(val1)+" and "+str(val2))
                delta = get_colour_distance(val1, val2)
                # LOGGER.debug("Colour distance " + str(delta))
                if delta > max_distance:
                    max_distance = delta
                if max_distance > porog:
                    return False

    return True


def calc_gradient_field(target_image: PIL.Image.Image) -> dict:
    """
    считает градиентное поля входного изображения (градиентом в точке называем сумму манхет.расстояний с ее соседями)
    :param target_image:
    :return: словарь. ключ - список [x,y] значение - градиент в точке [x,y]
    """
    target_image_pixels = target_image.load()

    gradient_field_dict = {}

    delta_x_lev = 0
    delta_y_verh = 0

    for x in range(0, target_image.width - 1):
        for y in range(0, target_image.height - 1):

            # LOGGER.debug("Calc gradient for " + str(x)+"x"+str(y))

            if x != target_image.width:
                delta_x_prav = get_colour_distance(target_image_pixels[x + 1, y], target_image_pixels[x, y])
            else:
                delta_x_prav = 0
            if y != target_image.height:
                delta_y_niz = get_colour_distance(target_image_pixels[x, y + 1], target_image_pixels[x, y])
            else:
                delta_y_niz = 0

            gradient_field_dict[(x, y)] = delta_y_niz + delta_x_lev + delta_x_prav + delta_y_verh

            delta_x_lev = delta_x_prav
            delta_y_verh = delta_y_niz

    LOGGER.debug("Gradient field dict: " + str(gradient_field_dict))

    return gradient_field_dict



import os
import PIL
from PIL import Image

from pho_mos.util import get_logger

LOGGER = get_logger("MOSAIC")


def open_as_pil_image(filename: str) -> PIL.Image:
    img = None
    try:
        img = Image.open(filename).convert('RGB')
    except Exception as exception:
        LOGGER.error("Error in _open: {0}".format(exception))
        if ".png" in filename:
            filename = filename.replace(".png", ".jpg")
            LOGGER.debug("Renamed: {0}".format(filename))
            img = Image.open(filename).convert('RGB')
    finally:
        # специально, чтобы заранее знать битые файлы
        assert img is not None
        return img


def square_img(image: PIL.Image,
               img_size: int = -1) -> PIL.Image:
    """
    Make image squared.

    :param img_size: размер результата. если не указан, соответствует минимальному габариту исходной
    :param image: Image
    :return: Image
    """
    assert image is not None
    try:
        width, height = image.size
        mean = abs(width - height) / 2.0
        if width > height:
            box = (mean, 0, width - mean, height)
        else:
            box = (0, mean, width, height - mean)
        image = image.crop(box=tuple(map(lambda x: int(x), box)))

        if img_size != -1:
            image = image.resize(size=(
                int(img_size), int(img_size)))

        return image
    except Exception as exception:
        print("Exception in preprocess_img: {0}".format(exception))
        return image


def resize_img(source_path: str,
               target_path: str = None,
               resize_factor: float = 64,
               is_save: bool = True) -> PIL.Image:
    """
    Меняем размер изорбажения в resize_factor раз.

    1. Makes img squared;
    2. Resize image;

    :param source_path: Union[os.path, str] - Path to the source img;
    :param target_path: Optional[Union[os.path, str]] - Path to save preprocessed img;
    :param resize_factor: int - во сколько раз увеличить изображение;
    :param is_return: bool - return result img or not;
    :param is_save: bool - save result img or not;
    :return:
    """
    try:

        image: PIL.Image = open_as_pil_image(source_path)

        if is_save:
            if not target_path:
                target_path = source_path

            if target_path == source_path and ".png" in target_path:
                os.remove(target_path)
                target_path = target_path.replace(".png", ".jpg")
                print(target_path)

        image = image.resize(size=(
            int(image.size[0]*resize_factor), int(image.size[1]*resize_factor))
                             )
        image = image.convert("RGB")
        if is_save:
            image.save(target_path)
        return image

    except Exception as exception:
        LOGGER.error("Exception in preprocess_img: {0}".format(exception))


def to_square_img(source_path: str,
               target_path: str = None,
               img_size: int = 64,
               is_save: bool = True) -> PIL.Image:
    """
    Превращает изображение в квадрат размера img_size

    1. Makes img squared;
    2. Resize image;

    :param source_path: Union[os.path, str] - Path to the source img;
    :param target_path: Optional[Union[os.path, str]] - Path to save preprocessed img;
    :param img_size: int - во сколько раз увеличить изображение;
    :param is_return: bool - return result img or not;
    :param is_save: bool - save result img or not;
    :return:
    """
    try:

        image: PIL.Image = open_as_pil_image(source_path)

        if is_save:
            if not target_path:
                target_path = source_path

            if target_path == source_path and ".png" in target_path:
                os.remove(target_path)
                target_path = target_path.replace(".png", ".jpg")
                print(target_path)

        image = square_img(image)
        image = image.resize(size=(img_size, img_size))
        if is_save:
            image.save(target_path)
        return image

    except Exception as exception:
        LOGGER.error("Exception in preprocess_img: {0}".format(exception))

'''
возвращает лист со строительными картинками нужного размера
'''
def batch_preprocess(source_dirname: str,
                     target_dirname: str = None,
                     img_size: int = 64,
                     is_save=True) -> list:
    """
    Image preprocessing by batch.

    Resize image;

    :param source_dirname: - source photos dir name;
    :param target_dirname: - target dir name to save photos;
    :param img_size: int - preprocessed img size;
    :param is_save: bool - save result imgs or not;
    :return: list of PIL.Image
    """
    try:
        print(source_dirname)

        if not os.path.exists(source_dirname):
            print("Warning: No such directory {0}".format(source_dirname))
            return

        if is_save:
            if not target_dirname:
                target_dirname = source_dirname
            else:
                if not os.path.exists(target_dirname):
                    os.mkdir(target_dirname)
        imgs = []

        for name in os.listdir(source_dirname):
            path = os.path.join(source_dirname, name)
            if is_save:
                targer_path = os.path.join(target_dirname, name)
            else:
                targer_path = ""
            img = to_square_img(path, targer_path, img_size, is_save=is_save)
            imgs.append(img)

        LOGGER.debug("Source images: "+str(imgs))

        return imgs

    except Exception as exception:
        LOGGER.error("Exception in batch_preprocess: {0}".format(exception))


import os

THIS_FILE_PATH = os.path.dirname(os.path.abspath(__file__))

import logging

main_component = "THEME_MOSAIC"
logging.basicConfig(
    format='%(levelname)s: [%(asctime)s.%(msecs)03d] {} %(name)s '
           '%(filename)s:%(funcName)s:%(lineno)s:  %(message)s'.format(main_component),
    datefmt='%Y-%m-%d %H:%M:%S', level='DEBUG')


def get_logger(TAG: str='') -> logging.Logger:
    return logging.getLogger(TAG)

# info - измениние состояния компанента системы
# debug - все потоковые события
# warning - обработанные ошибки (переподключение)
# error - Exceptions
# critical - не загрузились модули





class Point:
    def __init__(self, x: int, y:int):
        self.x = x
        self.y = y

class Rectangle:

    def __init__(self, width, height):
        self.width = width
        self.height = height

        # фигуры внутри текущего прямоугольника
        # ключ - Point левого верхего угла, значение - Rectangle
        self.inner_rectangles = {}

        # поле хранящее ссылку на изображение, которому соотвестствиет данный прямоугольник
        self.image: any = None


    def add_inner_rectangle(self, start_point: Point, rectangle):
        self.inner_rectangles[start_point] = rectangle

    # занята ли точка какой-то фигурой
    def is_occupied_point(self, point: Point) -> bool:
        for key_point, value_rect in self.inner_rectangles.items():
            if point.x>=key_point.x and point.x<=key_point.x+value_rect.width-1 \
                    and point.y>=key_point.y and point.y<=key_point.y+value_rect.height-1:
                return True
        return False






"""
A library for describing and applying affine transforms to PIL images.
SOURCE: https://gist.github.com/wchargin/d8eb0cbafc4d4479d004#file-transforms-py
"""
import numpy as np
import PIL.Image


class RGBTransform(object):
    """A description of an affine transformation to an RGB image.
    This class is immutable.
    Methods correspond to matrix left-multiplication/post-application:
    for example,
        RGBTransform().multiply_with(some_color).desaturate()
    describes a transformation where the multiplication takes place first.
    Use rgbt.applied_to(image) to return a converted copy of the given image.
    For example:
        grayish = RGBTransform.desaturate(factor=0.5).applied_to(some_image)
    """

    def __init__(self, matrix=None):
        self._matrix = matrix if matrix is not None else np.eye(4)

    def _then(self, operation):
        return RGBTransform(np.dot(_embed44(operation), self._matrix))

    def desaturate(self, factor=1.0, weights=(0.299, 0.587, 0.114)):
        """Desaturate an image by the given amount.
        A factor of 1.0 will make the image completely gray;
        a factor of 0.0 will leave the image unchanged.
        The weights represent the relative contributions of each channel.
        They should be a 1-by-3 array-like object (tuple, list, np.array).
        In most cases, their values should sum to 1.0
        (otherwise, the transformation will cause the image
        to get lighter or darker).
        """
        weights = _to_rgb(weights, "weights")

        # tile: [wr, wg, wb]  ==>  [[wr, wg, wb], [wr, wg, wb], [wr, wg, wb]]
        desaturated_component = factor * np.tile(weights, (3, 1))
        saturated_component = (1 - factor) * np.eye(3)
        operation = desaturated_component + saturated_component

        return self._then(operation)

    def multiply_with(self, base_color, factor=1.0):
        """Multiply an image by a constant base color.
        The base color should be a 1-by-3 array-like object
        representing an RGB color in [0, 255]^3 space.
        For example, to multiply with orange,
        the transformation
            RGBTransform().multiply_with((255, 127, 0))
        might be used.
        The factor controls the strength of the multiplication.
        A factor of 1.0 represents straight multiplication;
        other values will be linearly interpolated between
        the identity (0.0) and the straight multiplication (1.0).
        """
        component_vector = _to_rgb(base_color, "base_color") / 255.0
        new_component = factor * np.diag(component_vector)
        old_component = (1 - factor) * np.eye(3)
        operation = new_component + old_component

        return self._then(operation)

    def mix_with(self, base_color, factor=1.0):
        """Mix an image by a constant base color.
        The base color should be a 1-by-3 array-like object
        representing an RGB color in [0, 255]^3 space.
        For example, to mix with orange,
        the transformation
            RGBTransform().mix_with((255, 127, 0))
        might be used.
        The factor controls the strength of the color to be added.
        If the factor is 1.0, all pixels will be exactly the new color;
        if it is 0.0, the pixels will be unchanged.
        """
        base_color = _to_rgb(base_color, "base_color")
        operation = _embed44((1 - factor) * np.eye(3))
        operation[:3, 3] = factor * base_color

        return self._then(operation)

    def get_matrix(self):
        """Get the underlying 3-by-4 matrix for this affine transform."""
        return self._matrix[:3, :]

    def applied_to(self, image):
        """Apply this transformation to a copy of the given RGB* image.
        The image should be a PIL image with at least three channels.
        Specifically, the RGB and RGBA modes are both supported, but L is not.
        Any channels past the first three will pass through unchanged.
        The original image will not be modified;
        a new image of the same mode and dimensions will be returned.
        """

        # PIL.Image.convert wants the matrix as a flattened 12-tuple.
        # (The docs claim that they want a 16-tuple, but this is wrong;
        # cf. _imaging.c:767 in the PIL 1.1.7 source.)
        matrix = tuple(self.get_matrix().flatten())

        channel_names = image.getbands()
        channel_count = len(channel_names)
        if channel_count < 3:
            raise ValueError("Image must have at least three channels!")
        elif channel_count == 3:
            return image.convert('RGB', matrix)
        else:
            # Probably an RGBA image.
            # Operate on the first three channels (assuming RGB),
            # and tack any others back on at the end.
            channels = list(image.split())
            rgb = PIL.Image.merge('RGB', channels[:3])
            transformed = rgb.convert('RGB', matrix)
            new_channels = transformed.split()
            channels[:3] = new_channels
            return PIL.Image.merge(''.join(channel_names), channels)

    def applied_to_pixel(self, color):
        """Apply this transformation to a single RGB* pixel.
        In general, you want to apply a transformation to an entire image.
        But in the special case where you know that the image is all one color,
        you can save cycles by just applying the transformation to that color
        and then constructing an image of the desired size.
        For example, in the result of the following code,
        image1 and image2 should be identical:
            rgbt = create_some_rgb_tranform()
            white = (255, 255, 255)
            size = (100, 100)
            image1 = rgbt.applied_to(PIL.Image.new("RGB", size, white))
            image2 = PIL.Image.new("RGB", size, rgbt.applied_to_pixel(white))
        The construction of image2 will be faster for two reasons:
        first, only one PIL image is created; and
        second, the transformation is only applied once.
        The input must have at least three channels;
        the first three channels will be interpreted as RGB,
        and any other channels will pass through unchanged.
        To match the behavior of PIL,
        the values of the resulting pixel will be rounded (not truncated!)
        to the nearest whole number.
        """
        color = tuple(color)
        channel_count = len(color)
        extra_channels = tuple()
        if channel_count < 3:
            raise ValueError("Pixel must have at least three channels!")
        elif channel_count > 3:
            color, extra_channels = color[:3], color[3:]

        color_vector = np.array(color + (1, )).reshape(4, 1)
        result_vector = np.dot(self._matrix, color_vector)
        result = result_vector.flatten()[:3]

        full_result = tuple(result) + extra_channels
        rounded = tuple(int(round(x)) for x in full_result)

        return rounded


def _embed44(matrix):
    """Embed a 4-by-4 or smaller matrix in the upper-left of I_4."""
    result = np.eye(4)
    r, c = matrix.shape
    result[:r, :c] = matrix
    return result


def _to_rgb(thing, name="input"):
    """Convert an array-like object to a 1-by-3 numpy array, or fail."""
    thing = np.array(thing)
    assert thing.shape == (3, ), (
        "Expected %r to be a length-3 array-like object, but found shape %s" %
            (name, thing.shape))
    return thing


import setuptools

setuptools.setup(
    name="pho_mos",
    version="0.0.2",
    author="Vitaliy Rakitin",
    author_email="vitaliyrakitin@ya.ru",
    description="Photo mosaic",
    url="https://github.com/botanhuligan/pho_mos",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
